unit pmprint;

{ Copyright (c) Electro-Concept Mauricie, 1997-1998-1999-2000 }
{ TPlusMemoPrinter accessory component:

  Works with PlusMemo version 4.1 and later

  Dec. 22, 2000 release
    Added #12 as page break if found at the start of a paragraph
    Added PrintRange, PrintSelection and PrintFromDialog methods
    Added possibility to print only a range of pages from the preview window

  Oct. 21, 1999 release
    Added FooterFont, HeaderFont properties
    Header and footers now handle multiline text
    Zoom capability

  Initial release March 8, 1999 }

interface

uses
  WinTypes, WinProcs, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, PlusMemo;

type
  TPlusMemoPrinter = class(TComponent)
  private
    { Private declarations }
    fPlusMemo                 : TPlusMemo;

    fLeftMargin, fRightMargin,   { in inches }
    fTopMargin, fBottomMargin,
    fLineSpacing,
    fHeaderYPos, fFooterYPos  : Single;

    fHeaderAlign, fFooterAlign: TAlignment;
    fPlainText                : Boolean;

    fHeader, fFooter          : string;
    FHeaderFont, FFooterFont  : TFont;
    fPreviewTitle             : string;

    fPageCount, fCurrentPage  : Integer;
    fLinesPerPage             : Integer;
    fOnProgress               : TNotifyEvent;
    fOnPreview                : TNotifyEvent;
    fOnFormatPrint            : TNotifyEvent;

    { Internal working fields }
    fPrintAborted             : Boolean;
    fBufferMemo               : TPlusMemo;  { used as a copy of fPlusMemo
                                              to do the formatting and printing }
    fBufMemoFormatted         : Boolean;
    fSelection                : Boolean;    { True if fBufferMemo contains only the selected part of MemoToPrint }
    fPageBreaks               : TList;      { list of line numbers starting with a page break }
    fPageBreakLines           : TList;      { list of number of lines added for each page break }
    fTotalPageBreakLines      : Integer;    { total number of lines added due to page breaks }

    fLineHeight,
    fXSize, fYSize,
    fPageWidth, fPageHeight,
    fXStart, fYStart,
    fXPPI, fYPPI              : Integer;   { in pixels }

    procedure setPlusMemo(pm: TPlusMemo);
    procedure setLeftMargin(lm: Single);
    procedure setRightMargin(rm: Single);
    procedure setTopMargin(tm: Single);
    procedure setBottomMargin(bm: Single);
    procedure setLineSpacing(ls: Single);
    procedure setPlainText(pt: Boolean);
    procedure SetFooterFont(const Value: TFont);
    procedure SetHeaderFont(const Value: TFont);

  protected
    { Protected declarations }
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;

  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure AbortPrint;
    procedure FormatPrint;
    procedure Preview;
    procedure Print;
    procedure PrintFromDialog(PrintDialog: TPrintDialog);
    function  PrintLineToMemoLine(Line: Longint): Longint;
    procedure PrintRange(StartPage, StopPage: Integer);
    procedure PrintSelection;

    property PageCount   : Integer read fPageCount;
    property LinesPerPage: Integer read fLinesPerPage;
    property BufferMemo  : TPlusMemo read fBufferMemo;
    property CurrentPage : Integer read fCurrentPage;

    property PageHeight  : Integer read fPageHeight;
    property PageWidth   : Integer read fPageWidth;
    property PageXStart  : Integer read fXStart;
    property PageYStart  : Integer read fYStart;
    property PrintWidth  : Integer read fXSize;
    property PrintHeight : Integer read fYSize;
    property YPPI        : Integer read fYPPI;
    property XPPI        : Integer read fXPPI;
    property LineHeight  : Integer read fLineHeight;

  published
    { Published declarations }
    property MemoToPrint : TPlusMemo read fPlusMemo write setPlusMemo;

    property MarginLeft  : Single read fLeftMargin   write setLeftMargin;
    property MarginRight : Single read fRightMargin  write setRightMargin;
    property MarginTop   : Single read fTopMargin    write setTopMargin;
    property MarginBottom: Single read fBottomMargin write setBottomMargin;

    property LineSpacing : Single read fLineSpacing write setLineSpacing;
    property Header      : string read fHeader      write fHeader;
    property Footer      : string read fFooter      write fFooter;

    property HeaderYPos  : Single read fHeaderYPos  write fHeaderYPos;
    property FooterYPos  : Single read fFooterYPos  write fFooterYPos;

    property HeaderAlign : TAlignment read fHeaderAlign write fHeaderAlign;
    property FooterAlign : TAlignment read fFooterAlign write fFooterAlign;

    property HeaderFont  : TFont read FHeaderFont write SetHeaderFont;
    property FooterFont  : TFont read FFooterFont write SetFooterFont;

    property PlainText   : Boolean read fPlainText   write setPlainText;
    property PreviewTitle: string read fPreviewTitle write fPreviewTitle;

    property OnFormatPrint: TNotifyEvent read fOnFormatPrint write fOnFormatPrint;
    property OnProgress   : TNotifyEvent read fOnProgress write fOnProgress;
    property OnPreview    : TNotifyEvent read fOnPreview  write fOnPreview;

  end;

procedure Register;

implementation

uses StdCtrls, Printers, pmpprev;

constructor TPlusMemoPrinter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  fPageBreaks:= TList.Create;
  fPageBreakLines:= TList.Create;
  fLeftMargin:= 0.75;
  fRightMargin:= 0.5;
  fTopMargin:= 0.75;
  fBottomMargin:= 0.75;
  fLineSpacing:= - 1.5;
  fHeaderYPos:= 0.4;
  fFooterYPos:= 0.4;
  fHeaderFont:= TFont.Create;
  fFooterFont:= TFont.Create
end;

destructor TPlusMemoPrinter.Destroy;
begin
  fBufferMemo.Free;
  fPageBreaks.Free;
  fPageBreakLines.Free;
  fHeaderFont.Free;
  fFooterFont.Free;
  inherited Destroy
end;

procedure TPlusMemoPrinter.setPlusMemo(pm: TPlusMemo);
begin
  fPlusMemo:= pm;
  fBufferMemo.Free;
  fBufferMemo:= nil
end;

procedure TPlusMemoPrinter.setLeftMargin(lm: Single);
begin
  if lm<>fLeftMargin then
    begin
      fLeftMargin:= lm;
      fBufMemoFormatted:= False
    end
end;

procedure TPlusMemoPrinter.setRightMargin(rm: Single);
begin
  if rm<>fRightMargin then
    begin
      fRightMargin:= rm;
      fBufMemoFormatted:= False
    end
end;

procedure TPlusMemoPrinter.setTopMargin(tm: Single);
begin
  if tm<>fTopMargin then
    begin
      fTopMargin:= tm;
      setLineSpacing(fLineSpacing)   { update internal fields }
    end
end;

procedure TPlusMemoPrinter.setBottomMargin(bm: Single);
begin
  if bm<>fBottomMargin then
    begin
      fBottomMargin:= bm;
      setLineSpacing(fLineSpacing)  { update internal fields }
    end
end;

procedure ClearFPUExceptions;
  // clears pending FPU exceptions.
  // useful on wrong "invalid floating point operation" exceptions
  asm
    FNCLEX
  end;


procedure TPlusMemoPrinter.setLineSpacing(ls: Single);
var linespacingpix, i, j, add: Integer;
begin
  ClearFPUExceptions;
  fLineSpacing:= ls;
  if (fBufferMemo<>nil) and fBufMemoFormatted then
    begin
      if fLineSpacing<0 then linespacingpix:= -Round(fLineHeight*fLineSpacing)
                        else linespacingpix:= Round(fLineSpacing*fYPPI);

      if linespacingpix=0 then linespacingpix:=1;

      fLinesPerPage:= Trunc((fPageHeight-(MarginTop+MarginBottom)*fYPPI)/linespacingpix);
      if fLinesPerPage<=0 then fLinesPerPage:= 1;

      { add lines to make the page breaks }
      fTotalPageBreakLines:= 0;
      for i:= 0 to fPageBreaks.Count-1 do
        begin
          j:= Integer(fPageBreaks[i]) + fTotalPageBreakLines;
          add:= fLinesPerPage - j mod fLinesPerPage;
          Inc(fTotalPageBreakLines, add);
          fPageBreakLines[i]:= Pointer(add)
        end;

      fPageCount:= (fBufferMemo.LineCount + fTotalPageBreakLines) div fLinesPerPage + 1;
    end
end;

procedure TPlusMemoPrinter.SetFooterFont(const Value: TFont);
begin
  FFooterFont.Assign(Value);
end;

procedure TPlusMemoPrinter.SetHeaderFont(const Value: TFont);
begin
  FHeaderFont.Assign(Value);
end;

procedure TPlusMemoPrinter.setPlainText(pt: Boolean);
begin
  if pt<>fPlainText then
    begin
      fPlainText:= pt;
      fBufferMemo.Free;
      fBufferMemo:= nil
    end
end;

procedure TPlusMemoPrinter.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (Assigned(fPlusMemo)) and (AComponent = fPlusMemo) then
    begin
      fPlusMemo := nil;
      fBufferMemo.Free;
      fBufferMemo:= nil
    end
end;


procedure TPlusMemoPrinter.AbortPrint;
begin
  fPrintAborted:= True
end;

procedure TPlusMemoPrinter.Preview;
var fPreviewForm: TFrmPMPrintPreview;
begin
  fPreviewForm:= TFrmPMPrintPreview.Create(nil);
  with fPreviewForm do
    begin
      fPreviewForm.Caption:= fPreviewTitle;
      fPreviewForm.PMPrinter:= Self;
      if Assigned(fOnPreview) then OnPreview(fPreviewForm);
      ShowModal;
      Release
    end
end;

procedure TPlusMemoPrinter.Print;
begin
  if (fBufferMemo<>nil) and (fBufMemoFormatted) and (fSelection) then
    begin
      fBufferMemo.Lines.Assign(MemoToPrint.Lines);
      fBufMemoFormatted:= False
    end;
  fSelection:= False;
  FormatPrint;
  PrintRange(0, fPageCount-1)
end;

procedure TPlusMemoPrinter.PrintFromDialog(PrintDialog: TPrintDialog);
begin
  case PrintDialog.PrintRange of
    prAllPages: Print;
    prPageNums: PrintRange(PrintDialog.FromPage-1, PrintDialog.ToPage-1);
    prSelection: PrintSelection
    end
end;

procedure TPlusMemoPrinter.PrintSelection;
begin
  if (fBufferMemo<>nil) and (fBufMemoFormatted) and (not fSelection) then
    begin
      fBufferMemo.Lines.Text:= MemoToPrint.SelText;
      fBufMemoFormatted:= False
    end;
  fSelection:= True;
  FormatPrint;
  PrintRange(0, fPageCount-1)
end;

procedure TPlusMemoPrinter.PrintRange(StartPage, StopPage: Integer);
  { routines used for printing header and footer with page number embedded }
  function GetFooterRect: TRect;
    begin
      with Result do
        begin
          Bottom:= PageHeight - PageYStart;
          Top:= PageHeight - Round(FooterYPos*YPPI) - PageYStart;
          Left:= Round(MarginLeft*XPPI) - PageXStart;
          Right:= PageWidth - Round(MarginRight*XPPI) - PageXStart
        end
    end;

  function GetHeaderRect: TRect;
    begin
      with Result do
        begin
          Bottom:= Round(MarginTop*YPPI) - PageYStart;
          Top:= Round(HeaderYPos*YPPI) - PageYStart;
          Left:= Round(MarginLeft*XPPI) - PageXStart;
          Right:= PageWidth - Round(MarginRight*XPPI) - PageXStart
        end
    end;

  procedure PrintSpecial(PrRect: TRect; Align: TAlignment; s: string; F: TFont);
  const
    AlignToFlag: array[TAlignment] of Word = (DT_LEFT, DT_RIGHT, DT_CENTER);
  var
    pagepos: Integer;
  begin
    { put in page number }
    pagepos:= Pos('{p}', s);
    if pagepos>0 then
        s:= Copy(s, 1, pagepos-1) +
            IntToStr(fCurrentPage+1) +
            Copy(s, pagepos+3, Length(s)-pagepos-2);

    { print it }
    Printer.Canvas.Font:= F;
    DrawText(Printer.Canvas.Handle, PChar(s), -1, PrRect, AlignToFlag[Align] or DT_WORDBREAK)
  end;

var
  pagenb      : Integer;
  p           : TPoint;
  linespacingpix: Single;
  lastline, rline: Longint;

begin      { Print }
  ClearFPUExceptions;
  fPrintAborted:= False;
  FormatPrint;
  Printer.BeginDoc;

  try
    { determine the line spacing in pixels }
    if LineSpacing<0 then
      begin
        Printer.Canvas.Font:= fBufferMemo.Font;
        linespacingpix:= -Printer.Canvas.TextHeight('A')*LineSpacing
      end
    else linespacingpix:= LineSpacing*YPPI;

    if linespacingpix=0 then linespacingpix:=1;

    p.X:= Round(MarginLeft*XPPI-fXStart);
    p.Y:= Round(MarginTop*YPPI-fYStart);

    for pagenb:= StartPage to StopPage do
      begin
        fCurrentPage:= pagenb;
        if Assigned(Self.fOnProgress) then Self.OnProgress(Self);
        if fPrintAborted then Break;

        if pagenb>=fPageCount then Continue;
        if pagenb>StartPage then Printer.NewPage;
        if Header<>'' then PrintSpecial(GetHeaderRect, HeaderAlign, Header, HeaderFont);
        if Footer<>'' then PrintSpecial(GetFooterRect, FooterAlign, Footer, FooterFont);

        lastline:= (fCurrentPage+1)*LinesPerPage-1;
        repeat
          rline:= PrintLineToMemoLine(lastline);
          if rline=High(rline) then Dec(lastline)
        until (lastline<=0) or (rline<High(rline));

        fBufferMemo.DrawLines(PrintLineToMemoLine(fCurrentPage*LinesPerPage), rline,
                              Printer.Canvas, p, linespacingpix);
      end;


  finally
    Printer.EndDoc;
    fCurrentPage:= -1;
    if Assigned(Self.fOnProgress) then Self.OnProgress(Self);
  end;
end;  { PrintRange }

{$IFNDEF WIN32}
const PHYSICALWIDTH = HORZRES;
      PHYSICALHEIGHT= VERTRES;

{$ENDIF}

type TprintMemo = class(TPlusMemo);  { to expose fParagraphs property }

procedure TPlusMemoPrinter.FormatPrint;
var
  pdevice, pdriver, pport: PChar;
  pdevmodehand: THandle;
  pdevmode    : Pointer;
  sBufCanvas  : TCanvas;   { a canvas whose handle is the printer device context }
  dc          : THandle;
  i           : Longint;
  sp          : pParInfo;

begin
  if fBufferMemo=nil then
    begin
      fBufMemoFormatted:= False;
      fBufferMemo:= TPlusMemo.Create(nil);

      if MemoToPrint<>nil then
        with MemoToPrint do
          begin
            fBufferMemo.Alignment         := Alignment;
            fBufferMemo.AltFont           := AltFont;
            fBufferMemo.ApplyKeywords     := ApplyKeywords;
            fBufferMemo.ApplyStartStopKeys:= ApplyStartStopKeys;
            fBufferMemo.Delimiters        := Delimiters;
            fBufferMemo.Justified         := Justified;
            fBufferMemo.Font              := Font;
            fBufferMemo.LeftMargin        := 0;
            fBufferMemo.RightMargin       := 0;
            fBufferMemo.TabStops          := TabStops;
            fBufferMemo.UpdateMode        := umOnNeed;
            fBufferMemo.Visible           := False;
            fBufferMemo.WordWrap          := WordWrap;
            fBufferMemo.Keywords.Free;
            fBufferMemo.StartStopKeys.Free;
            fBufferMemo.Keywords          := nil;
            fBufferMemo.StartStopKeys     := nil;
            fBufferMemo.ScrollBars        := ssNone;

            if PlainText then
              with fBufferMemo do
                begin
                  ApplyKeywords      := False;
                  ApplyStartStopKeys := False
                end
            else
              begin
                fBufferMemo.ApplyKeywords      := ApplyKeywords;
                fBufferMemo.ApplyStartStopKeys := ApplyStartStopKeys
              end;

            if fSelection then fBufferMemo.Lines.Text:= SelText
                          else fBufferMemo.Lines.Assign(Lines);
          end;      { with MemoToPrint }

    end;  { fBufferMemo=nil }

  if not fBufMemoFormatted then
    begin
      fPageBreaks.Clear;
      fPageBreakLines.Clear;
      fTotalPageBreakLines:= 0;

      fBufferMemo.Parent:= MemoToPrint;   { TPlusMemo needs a handle to perform
                                            its FormatText function.  This requirement
                                            will be removed in version 4.1g and later }
      fBufferMemo.Keywords:= MemoToPrint.Keywords;
      fBufferMemo.StartStopKeys:= MemoToPrint.StartStopKeys;
      if not PlainText then fBufferMemo.Highlighter:= MemoToPrint.Highlighter;
      if Assigned(fOnFormatPrint) then fOnFormatPrint(fBufferMemo);
      fBufferMemo.ReApplyKeywords;

      { get the page characteristics }
      with Printer do
        begin
          fPageWidth := GetDeviceCaps(Handle, PHYSICALWIDTH);
          fPageHeight:= GetDeviceCaps(Handle, PHYSICALHEIGHT);
          {$IFDEF WIN32}
          fXStart    := GetDeviceCaps(Handle, PHYSICALOFFSETX);
          fYStart    := GetDeviceCaps(Handle, PHYSICALOFFSETY);
          {$ELSE}
          fXStart    := 0;
          fYStart    := 0;
          {$ENDIF}
          fXSize     := GetDeviceCaps(Handle, HORZRES);
          fYSize     := GetDeviceCaps(Handle, VERTRES);
          fXPPI      := GetDeviceCaps(Handle, LOGPIXELSX);
          fYPPI      := GetDeviceCaps(Handle, LOGPIXELSY);

        end;

      { get the printer device context in sBufCanvas.Handle }
      pdevice:= StrAlloc(1000);
      pdriver:= StrAlloc(1000);
      pport  := StrAlloc(1000);

      Printer.GetPrinter(pdevice, pdriver, pport, pdevmodehand);
      pdevmode:= GlobalLock(pdevmodehand);

      sBufCanvas:= TCanvas.Create;
      sBufCanvas.Handle:= CreateDC(pdriver, pdevice, pport, pdevmode);
      sBufCanvas.Font.PixelsPerInch:= fYPPI;
      sBufCanvas.Font:= fBufferMemo.Font;
      fLineHeight:= sBufCanvas.TextHeight('A');

      fBufferMemo.HandleNeeded;
      fBufferMemo.OnProgress:= OnFormatPrint;
      fBufferMemo.FormatText(sBufCanvas, fPageWidth-Round(MarginLeft*fXPPI)-Round(MarginRight*fXPPI));
      fBufferMemo.Keywords:= nil;
      fBufferMemo.StartStopKeys:= nil;
      fBufferMemo.Highlighter:= nil;

      dc:= sBufCanvas.Handle;
      sBufCanvas.Handle:= 0;
      DeleteDC(DC);

      StrDispose(pdevice);
      StrDispose(pdriver);
      StrDispose(pport);
      sBufCanvas.Free;
      GlobalUnlock(pdevmodehand);
      fBufMemoFormatted:= True;
      fBufferMemo.Parent:= nil;

      { determine the location of all page breaks }
      for i:= 0 to fBufferMemo.ParagraphCount-1 do
        begin
          sp:= TPrintMemo(fBufferMemo).fParagraphs.ParPointers[i];
          if (sp^.Length>=1) and (sp^.Text[0]=#12) then fPageBreaks.Add(Pointer(sp.StartLine))
        end;
      fPageBreakLines.Count:= fPageBreaks.Count

    end;     { not fBufMemoFormatted }

  setLineSpacing(fLineSpacing)   { update other internal fields }
end;    { FormatPrint }

function TPlusMemoPrinter.PrintLineToMemoLine(Line: Longint): Longint;
var i: Integer;
begin
  Result:= Line;
  i:= 0;
  while (i<fPageBreaks.Count) and (Result>=Integer(fPageBreaks[i])) do
    begin
      Dec(Result, Integer(fPageBreakLines[i]));
      if Result<Integer(fPageBreaks[i]) then
        begin
          Result:= High(Result);
          Break
        end;
      Inc(i)
    end
end;


procedure Register;
begin
  RegisterComponents('ECM', [TPlusMemoPrinter]);
end;

end.
